(function() 
 {

     Ext.define('model_wiz.helpers.Stream',
                function() {
                    var topics = {};
                    var subUid = 0;

                    // getTopic :: String -> Topic [a]
                    function getTopic(t) {
                        return (topics[t] || []);
                    }

                    // trigger :: String -> ([Arguments] || Arguments) -> Optional ([Ids] || Id) -> Optional Object -> [Results]
                    function trigger(topic, args, ids, context) {
                        var a = [].concat(args);
                        var i = [].concat(ids);
                        return getTopic(topic)
                        .reduce(function(acc, o) {
                            return (! ids || i.indexOf(o.id) > -1)
                            ? acc.concat(
                                {
                                    topic: topic,
                                    id: o.id,
                                    result:  o.func.apply((context || o), a)
                                }
                            )
                            : acc;
                        }, []);
                    }

                    var sleeping = false;

                    // sub :: String -> f -> Optional Id -> String
                    function sub(topic, fn, id) {
                        topics[topic] = getTopic(topic).concat(
                            {
                                id: (id || subUid++).toString(),
                                func: fn,
                                // then :: g -> h -> g f
                                then: function(g) {
                                    var f = this.func;
                                    this.func = function() {
                                        var a = Array.prototype.slice.call(arguments,0);
                                        return g.call(this, f.apply(this, a));
                                    };
                                    return this;
                                },
                                sleep: function(time) {
                                    var f = this.func;
                                    this.func = function() {
                                        var a = Array.prototype.slice.call(arguments,0);
                                        var me = this;
                                        clearTimeout(sleeping);
                                        sleeping = setTimeout(function() {
                                            return f.apply(me, a);
                                        }, time);
                                    };
                                    return this;
                                }
                            }
                        );
                        return model_wiz.helpers.Utils.last(topics[topic]);
                    }

                    // unsub :: String -> Id -> [a]
                    function unsub(topic, id) {
                        return (topics[topic] !== null) || (topics[topic] !== undefined)
                        ? (topics[topic] = getTopic(topic).reduce(function(acc, o) {
                            return (! id) || (id.some(function(curr) { return curr === o.id; }))
                            ? acc
                            : acc.concat(o);
                        }, []), topics[topic])
                        : {};
                    }

                    // getTopicIds :: String -> [Ids]
                    function getTopicIds(topic) {
                        return getTopic(topic).map(function(t) {
                            return t.id;
                        });
                    }

                    // trace :: String -> (f -> _ -> f a)
                    function trace(tag) {
                        return function(a) {
                            console.log(tag + ': ', a);
                            return a;
                        };
                    }

                    // bind :: f -> a -> g -> (g -> _ -> f a)
                    function bind(f) {
                        var fa = Array.prototype.slice.call(arguments,1);
                        var me = this;
                        return function() {
                            var a = Array.prototype.slice.call(arguments,0);
                            return f.apply(me, fa.concat(a));
                        };
                    }

                    function getTopics() {
                        return Ext.apply(topics, {});
                    }

                    // Stream interface
                    return {
                        statics: {
                            publish: trigger,
                            subscribe: sub,
                            unsubscribe: unsub,
                            getTopicIds: getTopicIds,
                            trace: trace,
                            bind: bind,
                            getTopics: getTopics
                        }
                    };
                }
               );

     Ext.define('model_wiz.helpers.Utils',
                function() {
                    // curry :: f -> g
                    function curry(fx) {
                        var arity = fx.length;
                        return function f1() {
                            var args = Array.prototype.slice.call(arguments, 0);
                            return (args.length >= arity)
                            ? fx.apply(null, args)
                            : function() {
                                var args2 = Array.prototype.slice.call(arguments, 0);
                                return f1.apply(null, args.concat(args2));
                            };
                        };
                    }

                    // cloneUt :: a -> b
                    function cloneUt(obj) {
                        if (null === obj || "object" != typeof obj) return obj;
                        var copy = obj.constructor();
                        for (var attr in obj) {
                            if (obj.hasOwnProperty(attr)) copy[attr] = obj[attr];

                        }
                        return copy;
                    }

                    // reduceUt :: [a] -> String -> Boolean -> b
                    function reduceUt(arr, key, createPropertyValue) {
                        return arr.reduce(function(prev, curr) {
                            prev[curr[key].toLowerCase()] = cloneUt(curr);
                            if (createPropertyValue === true) {
                                prev[curr[key].toLowerCase()].value = null;
                            }
                            return prev;
                        }, {});
                    }

                    // head :: [a] -> a
                    function head(xs) {
                        return xs[0];
                    }

                    // tail :: [a] -> [a]
                    function tail(xs) {
                        return xs.slice(1);
                    }

                    // last :: [a] -> a
                    function last(arr) {
                        return arr[arr.length - 1];
                    }

                    // matchClassName :: Object, Object -> Array
                    function matchClassName(t, r) {
                        return t.className.match(r);
                    }

                    // pluck :: String -> Object -> a
                    var pluck = curry(function(keys, obj) {
                        var k = [].concat(keys);
                        return k.length > 0
                        ? pluck(k.slice(1), obj[head(k)])
                        : obj;
                    });

                    // catchOpposite :: a -> Boolean
                    function catchOpposite(a) {
                        return ! (!!a);
                    }

                    // id :: a -> a
                    function id(a) {
                        return a;
                    }

                    // flatten :: [[a]] -> [a]
                    function flatten(xxs) {
                        return xxs.reduce(function(acc, xs) {
                            return acc.concat(xs);
                        }, []);
                    }

                    // choseAct :: String -> String -> f -> a -> bind(g, a)
                    function choseAct(reg, el, operations, context) {
                        var target = matchClassName(el.target, reg) || [];
                        var doestnExistMsg = console.log.bind(console, 'There isnt a controller function to this action');
                        var fn = operations[target[1]] || doestnExistMsg;
                        return fn.bind(context);
                    }

                    Ext.MessageBox.buttonText = {
                        yes    : Ext.locale.translate('Yes'),
                        no     : Ext.locale.translate('No')
                    };

                    // confirmAct :: String -> f -> a -> b
                    var confirmAct = function(tag, op, me) {
                        return function(value) {
                            return Ext.Msg.confirm(Ext.locale.translate('Confirm change'), tag, function(res) {
                                return res === 'yes' ? op.call(me, value) : false;
                            });
                        };
                    };

                    // add :: Number -> Number -> Number
                    function add(a,b) {
                        return a + b;
                    }

                    // and :: a -> b -> Boolean
                    function and(a,b) {
                        return a && b;
                    }

                    // or :: a -> b -> Boolean
                    function or(a,b) {
                        return a || b;
                    }

                    // True :: _ -> Boolean
                    function True() {
                        return true;
                    }

                    // False :: _ -> Boolean
                    function False() {
                        return false;
                    }

                    // fromNullable :: a -> Either null a
                    function fromNullable(x) {
                        return (x === null) || (x === undefined)
                        ? model_wiz.Functor.Left(null)
                        : model_wiz.Functor.Right(x);
                    }

                    // fromEmpty :: a -> Either null a
                    function fromEmpty(x) {
                        return fromNullable(x)
                        .chain(function(x) {
                            return ((x === "") || (x.length === 0))
                            ? model_wiz.Functor.Left.of(x)
                            : model_wiz.Functor.Right.of(x);
                        });
                    }

                    // fromFalse :: a -> Either Boolean Boolean
                    var fromFalse = function(x) {
                        return (x === false)
                        ? model_wiz.Functor.Right(x)
                        : model_wiz.Functor.Left(x);
                    };

                    // fromTrue :: a -> Boolean
                    var fromTrue = function(x) {
                        return (x === true)
                        ? model_wiz.Functor.Right(x)
                        : model_wiz.Functor.Left(x);
                    };

                    // bind :: f -> g
                    var bind = function(f) {
                        var args = Array.prototype.slice.call(arguments, 1);
                        return function() {
                            var args2 = Array.prototype.slice.call(arguments);
                            return f.apply(null, args.concat(args2));
                        };
                    };

                    // glueContext :: a -> f -> g
                    var glueContext = curry(function(ctx, f) {
                        return tryCatch(function() {
                            return f.bind(ctx);
                        });
                    });

                    // compose :: ...f -> g
                    var compose = function() {
                        var fns = Array.prototype.slice.call(arguments);
                        return function(data) {
                            return fns.reduceRight(function(val, f) {
                                return f(val);
                            }, data);
                        };
                    };

                    // match :: Regex -> String -> Either String String
                    var match = curry(function(regex, str) {
                        return str.toString().match(regex)
                        ? model_wiz.Functor.Right(str)
                        : model_wiz.Functor.Left(str);
                    });

                    // notMatch :: Regex -> String -> Either String String
                    var notMatch = curry(function(regex, str) {
                        return match(regex, str)
                        .fold(model_wiz.Functor.Right.of,
                              model_wiz.Functor.Left.of);
                    });

                    // flip :: f -> g
                    var flip = function(fn) {
                        return function () {
                            for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                                args[_key] = arguments[_key];
                            }
                            args = args.slice(1, 2).concat(args.slice(0, 1)).concat(args.slice(2));
                            return fn.apply(null, args);
                        };
                    };

                    // execWithContext :: a -> f -> f apply a
                    var execWithContext = curry(function(context, f) {
                        var args = Array.prototype.slice.call(arguments, 2);
                        return f.apply(context, args);
                    });

                    // lowercase :: String -> String
                    var lowercase = function(str) {
                        return str.toLowerCase();
                    };

                    // uppercase :: String -> String
                    var uppercase = function(str) {
                        return str.toUpperCase();
                    };

                    // keys :: Object -> [String]
                    var keys = function(a) {
                        return (Object.keys(a) || []);
                    };

                    // find :: f -> [a] -> a
                    var find = function(f, arr) {
                        return arr.reduce(function(acc, curr) {
                            return (acc === null) && (f(curr))
                            ? curr
                            : acc;
                        }, null);
                    };

                    var findRecord = curry(function(store, key, val) {
                        return model_wiz.Functor.Right.of(curry(
                            function(k,v,s) { return s.findRecord(k,v,0,false,false,true); }
                        ))
                        .ap(fromEmpty(key))
                        .ap(fromEmpty(val))
                        .ap(fromNullable(Ext.getStore(store)))
                        .chain(fromNullable);
                    });

                    // map :: Functor f => g -> f -> f g -> h
                    var map = curry(function(f, functor) {
                        return functor.map(f);
                    });

                    var chain = curry(function(f, functor) {
                        return functor.chain(f);
                    });

                    // listMap :: Functor f => g -> f -> f g -> h
                    var listMap = curry(function(f, xs) {
                        return xs.map(function(curr) { return curr.map(f); });
                    });

                    // listFold
                    var listFold = function(xs) {
                        return xs.reduce(function(acc, curr) {
                            return curr.fold(function() { return acc; },
                                             concat(acc));
                        }, []);
                    };

                    // hasProperty :: Key -> Object -> Boolean
                    var hasProperty = curry(function(key, a) {
                        return a.hasOwnProperty(key);
                    });

                    // noProperty :: Key -> Object -> Boolean
                    var noProperty = curry(function(key, a) {
                        return a.hasOwnProperty(key) === false;
                    });

                    // isPropertyEmpty :: Key -> Object -> Boolean
                    var isPropertyEmpty = curry(function(key, a) {
                        return a.hasOwnProperty(key) && a[key] === null || a[key] === undefined;
                    });

                    // concat :: a -> b -> c
                    var concat = curry(function(a, b) {
                        return a.concat(b);
                    });

                    // findRecord :: Ext Store -> String -> a -> Either Null Record
                    var findRecord = curry(function(store, key, val) {
                        return model_wiz.Functor.Right.of(
                            curry(function(k, v, s) {
                                return s.findRecord(k, v, 0, false, false, true);
                            })
                        )
                        .ap(fromEmpty(key))
                        .ap(fromEmpty(val))
                        .ap(fromNullable(Ext.getStore(store)))
                        .chain(fromNullable);
                    });

                    var filterRecord = curry(function(store, key, val) {
                        return model_wiz.Functor.Right.of(
                            curry(function(d, k, v) {
                                return d.filter(function(curr) {
                                    return (curr.get(k) === v);
                                });
                            })
                        )
                        .ap(compose(fromNullable, pluck(['data', 'items']), Ext.getStore) (store))
                        .ap(fromNullable(key))
                        .ap(fromNullable(val));
                    });

                    var filterNotDuplicatedInStore = curry(function(store, testKey, xs) {
                        return model_wiz.Functor.Right.of(curry(
                            function(s,k,xs) {
                                return xs.reduce(function(acc, curr) {
                                    return compose(existsSome(curr[k], k), map(pluck('data')), pluck(['data', 'items'])) (s)
                                    ? acc
                                    : acc.concat(curr);
                                }, []);
                            }
                        ))
                        .ap(getStore(store))
                        .ap(fromNullable(testKey))
                        .ap(fromNullable(xs));
                    });

                    var addRecord = curry(function(store, xs) {
                        return model_wiz.Functor.Right.of(curry(
                            function(s,xs) {
                                return tryCatch(s.add.bind(s, xs)).fold(id, id);
                            }
                        ))
                        .ap(getStore(store))
                        .ap(fromNullable(xs));
                    });

                    var removeAllRecords = curry(
                        function(store, val) {
                            return fromNullable(store)
                            .chain(getStore)
                            .chain(function(s) {
                                return tryCatch(function() {
                                    s.removeAll();
                                    return val;
                                });
                            });
                        }
                    );

                    var getStore = function(store) {
                        return compose(chain(fromNullable), tryCatch) (Ext.getStore.bind(Ext, store))
                        .bimap(function(e) { return 'Store ' + store + ' result in a error: ' + e; },
                               id);
                    };

                    var getController = function(controller) {
                        return compose(chain(fromNullable), tryCatch) (model_wiz.app.getController.bind(model_wiz.app, controller))
                        .bimap(function(e) { return 'Controller ' + controller + ' result in a error: ' + e; },
                               id);
                    };

                    // headObj :: a -> b
                    var headObj = function(a) {
                        return a[compose(head, Object.keys) (a)];
                    };

                    // trace :: String -> a -> a
                    var trace = curry(function(tag, a) {
                        console.log('# Trace #\n' + tag + ': ', a);
                        return a;
                    });

                    // exec :: f -> a
                    var exec = function(f) { return f(); };

                    // split :: String -> String -> [String]
                    var split = curry(function(reg, str) {
                        return str.split(reg);
                    });

                    var slice = curry(function(start, arr) {
                        return arr.slice(start);
                    });

                    // getStoreData :: String -> [Object]
                    var getStoreData = function(store) {
                        return Ext.getStore(store);
                    };

                    // setRecord
                    var setRecord = curry(function(rec, fieldname, value) {
                        return rec.set(fieldname, value);
                    });

                    var safeProp = curry(function(p, o) {
                        return compose(fromNullable, pluck(p)) (o);
                    });

                    var safePluck = curry(function(keys, o) {
                        return [].concat(keys).reduce(function(acc, curr) {
                            return acc.chain(safeProp(curr));
                        }, fromNullable(o));
                    });

                    var ifElse = curry(function(predicate, f, g, val) {
                        return predicate(val) ? f(val) : g(val);
                    });

                    var isPropertyTrue = curry(function(p, o) {
                        return hasProperty(p, o) && o[p] === true;
                    });

                    var isPropertyFalse = curry(function(p, o) {
                        return hasProperty(p, o) && (o[p] === false || o[p] === null || o[p] === undefined);
                    });

                    var extractArgs = curry(function(start, xs) {
                        return Array.prototype.slice.call(xs, start);
                    });

                    var tryCatch = function(f) {
                        try {
                            return model_wiz.Functor.Right.of(f());
                        }
                        catch(err) {
                            return model_wiz.Functor.Left.of(err);
                        }
                    };

                    // bindContext :: a -> f -> g
                    var bindContext = curry(function(ctx, f) {
                        return f.bind(ctx);
                    });

                    var existsSome = curry(function(val, key, xs) {
                        return xs.some(function(curr) { return curr[key] === val; });
                    });

                    var extClone = function(o) {
                        return Ext.clone(o);
                    };

                    var apply = curry(function(o2, o1) {
                        return Ext.apply(o1, o2);
                    });

                    var safeApply = curry(function(o1,o2) {
                        return model_wiz.Functor.Right.of(curry(function(x,y,f) {
                            return tryCatch(function() { return f(x,y); });
                        }))
                        .ap(fromNullable(o1))
                        .ap(fromNullable(o2))
                        .ap(compose(map(bindContext(Ext)), safePluck('apply')) (Ext))
                        .chain(id);
                    });

                    var filter = curry(function(f, xs) {
                        return xs.filter(f);
                    });

                    var cleanStore = curry(function(f, s) {
                        return model_wiz.Functor.Right.of(curry(
                            function(records, store) {
                                return compose(store.remove.bind(store), filter(f)) (records);
                            }
                        ))
                        .ap(compose(safePluck(['data','items']), getStore) (s))
                        .ap(getStore(s));
                    });

                    var setRecord = curry(function(key, value, record) {
                        fromNullable(curry(function(k,v,f) {
                            return tryCatch(function() {
                                return f(k,v);
                            });
                        }))
                        .ap(fromNullable(key))
                        .ap(fromNullable(value))
                        .ap(compose(map(bindContext(record)), safePluck('set')) (record))
                        .chain(id);
                    });

                    var safeClone = function(obj) {
                        return fromNullable(Ext)
                        .chain(safePluck('clone'))
                        .map(bindContext(Ext))
                        .map(curry(function(f,o) {
                            return tryCatch(function() { return f(o); });
                        }))
                        .ap(fromNullable(obj))
                        .chain(id);
                    };

                    var getRecordFn = curry(function(f,r) {
                        return fromNullable(r)
                        .chain(safePluck(f))
                        .map(bindContext(r));
                    });

                    var getObjFn = getRecordFn;

                    var logError = curry(function(prefix, err) {
                        return curry(function(p,e) {
                            return tryCatch(function() {
                                console.error('# ERR LOG # '.concat(prefix).concat('\n######## Error Message ########\n'), err);
                                return err;
                            });
                        })
                        .ap(fromNullable(prefix))
                        .ap(fromNullable(err))
                        .chain(id);
                    });

                    var resultIn = curry(function(a,_) {return a;});

                    var isEquals = curry(function(a,b,c) {
                        return model_wiz.Functor.Right.of(curry(function(x,y) {
                            return (x === y)
                            ? model_wiz.Functor.Right.of(c)
                            : model_wiz.Functor.Left.of(c);
                        }));
                    });

                    var log = curry(function(log, val) {
                        return fromNullable(log)
                        .chain(function(l) { return tryCatch(function() { console.log('# Log #\n', l); }); })
                        .map(resultIn(val));
                    });

                    return {
                        statics: {
                            clone			: cloneUt,
                            extClone		: extClone,
                            reduce			: reduceUt,
                            head			: head,
                            tail			: tail,
                            last			: last,
                            choseAction		: choseAct,
                            confirmAction	: confirmAct,
                            catchOpposite	: catchOpposite,
                            pluck			: pluck,
                            identity		: id,
                            flatten			: flatten,
                            add				: add,
                            fromFalse		: fromFalse,
                            fromTrue		: fromTrue,
                            and				: and,
                            or				: or,
                            curry			: curry,
                            True			: True,
                            False			: False,
                            fromNullable	: fromNullable,
                            bind			: bind,
                            compose			: compose,
                            fromEmpty		: fromEmpty,
                            match			: match,
                            notMatch		: notMatch,
                            flip 			: flip,
                            execWithContext	: execWithContext,
                            lowercase		: lowercase,
                            uppercase		: uppercase,
                            keys			: keys,
                            find			: find,
                            map 			: map,
                            chain 			: chain,
                            hasProperty		: hasProperty,
                            concat 			: concat,
                            findRecord		: findRecord,
                            headObj			: headObj,
                            trace			: trace,
                            exec			: exec,
                            split 			: split,
                            listMap			: listMap,
                            listFold 		: listFold,
                            setRecord		: setRecord,
                            safeProp		: safeProp,
                            ifElse 			: ifElse,
                            isPropertyTrue	: isPropertyTrue,
                            filterRecord	: filterRecord,
                            filterNotDuplicatedInStore : filterNotDuplicatedInStore,
                            addRecord		: addRecord,
                            tryCatch		: tryCatch,
                            bindContext		: bindContext,
                            existsSome		: existsSome,
                            getStore		: getStore,
                            apply			: apply,
                            getController 	: getController,
                            removeAllRecords : removeAllRecords,
                            safePluck		: safePluck,
                            slice 			: slice,
                            cleanStore		: cleanStore,
                            filter			: filter,
                            isPropertyFalse	: isPropertyFalse,
                            noProperty		: noProperty,
                            isPropertyEmpty : isPropertyEmpty,
                            glueContext		: glueContext,
                            setRecord		: setRecord,
                            safeClone 		: safeClone,
                            safeApply		: safeApply,
                            getRecordFn		: getRecordFn,
                            logError		: logError,
                            resultIn		: resultIn,
                            isEquals		: isEquals,
                            log				: log,
                            getObjFn		: getObjFn
                        }
                    };
                }
               );

     Ext.define('model_wiz.Functor', function() {
         var Box = function(x) {
             return {
                 x: x,
                 map: function(f) {
                     return Box.of(f(x));
                 },
                 ap: function(other) {
                     return other.map(x);
                 },
                 chain: function(f) {
                     return f(x);
                 },
                 fold: function(f) {
                     return f(x);
                 }
             };
         };

         Box.of = function(x) { return new Box(x); };

         var Left = function(x) {
             return {
                 x: x,
                 map: function(f) {
                     return Left.of(x);
                 },
                 ap: function(other) {
                     return Left.of(x);
                 },
                 chain: function(f) {
                     return Left.of(x);
                 },
                 bimap: function(f, g) {
                     return Left.of(f(x));
                 },
                 mapBoth: function(f) {
                     return Left.of(f(x));
                 },
                 leftMap: function(f) {
                     return f(x);
                 },
                 fold: function(f, g) {
                     return f(x);
                 }
             };
         };

         Left.of = function(x) { return new Left(x); };

         var Right = function(x) {
             return {
                 x: x,
                 map: function(f) {
                     return Right.of(f(x));
                 },
                 ap: function(other) {
                     return other.map(x);
                 },
                 chain: function(f) {
                     return f(x);
                 },
                 bimap: function(f, g) {
                     return Right.of(g(x));
                 },
                 mapBoth: function(f) {
                     return Right.of(f(x));
                 },
                 leftMap: function(f) {
                     return Right.of(x);
                 },
                 fold: function(f, g) {
                     return g(x);
                 }
             };
         };

         Right.of = function(x) { return new Right(x); };

         return {
             statics: {
                 Box: Box, Left: Left, Right: Right
             }
         };
     });
 })();
